package org.artifactory.queue.register;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.queue.QueueConfiguration;
import org.artifactory.queue.services.PersistentQueue;
import org.artifactory.queue.services.PersistentQueueImpl;
import org.artifactory.queue.services.QueueOperations;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.db.event.entity.NodeEventCursorType;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.config.diff.DataDiff;
import org.springframework.stereotype.Component;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.stream.Collectors;

import static org.artifactory.storage.db.event.entity.NodeEventCursorType.getEventIdLogTypes;

/**
 * The registration service is responsible to create and remove all PersistentQueue's instances.
 * an instance of PersistentQueue is being created for each repoKey+cursorType+variant<br>
 * e.g:<br>
 * repoNames = my-local-maven1 ,my-local-maven2 ,my-virtual-maven1
 * cursorType = Incremental-index, replication(var1,var2)
 * will create this 9 PersistentQueues instances:
 * -  my-local-maven1-incremental-index-queue
 * -  my-local-maven2-incremental-index-queue
 * -  my-virtual-maven1-incremental-index-queue
 * -  my-local-maven1-replication-var1-queue
 * -  my-local-maven1-replication-var2-queue
 * -  my-local-maven2-replication-var1-queue
 * -  my-local-maven2-replication-var2-queue
 * -  my-virtual-maven1-replication-var1--queue
 * -  my-virtual-maven1-replication-var2--queue
 *
 * The creation process is being done on start up (init method) and whenever a new repository is added
 * (registerRepository method) - by collecting all QueueConfigurationRegistrar implementations.
 * and by application code using the QueueService.register method which accept a single
 * QueueConfiguration without the need to create a Registrar implementation.
 *
 * the removal of a queue is being done whenever a repository is deleted (deRegisterRepository method)
 *
 * @author Avishay Halpren
 */
@Slf4j
@Component
@Reloadable(beanClass = PersistentQueueRegistrationService.class, initAfter = {InternalRepositoryService.class},
        listenOn = {
                CentralConfigKey.localRepositoriesMap,
                CentralConfigKey.federatedRepositoriesMap,
                CentralConfigKey.remoteRepositoriesMap,
                CentralConfigKey.virtualRepositoriesMap,
        })
public class PersistentQueueRegistrationServiceImpl implements PersistentQueueRegistrationService {

    /**
     * Stores all the queues created by this service.
     * the key is repoKey + cursorType for multiple (variant based) queues
     */
    private final Multimap<RepoAndCursor, PersistentQueue> queues =
            Multimaps.synchronizedListMultimap(ArrayListMultimap.create());
    /**
     * this lock defends modifications of queues multimap which is not thread safe.
     */
    private final ReentrantReadWriteLock readWriteLock = new ReentrantReadWriteLock();

    private final List<QueueConfigurationRegistrar> registrars;
    private final RepositoryService repositoryService;

    public PersistentQueueRegistrationServiceImpl(List<QueueConfigurationRegistrar> registrars,
            RepositoryService repositoryService) {
        this.registrars = registrars;
        this.repositoryService = repositoryService;
    }

    @Override
    public void init() {
        //register all existing repositories on startup
        repositoryService.getAllRepoKeys().forEach(repoKey -> registerRepository(repoKey, false));
    }

    @Override
    public void reload(CentralConfigDescriptor oldDescriptor, List<DataDiff<?>> configDiff) {
        //This method does not support update.
        for (DataDiff<?> diff : configDiff) {
            // new repo
            if (diff.getNewValue() instanceof RepoDescriptor && diff.getOldValue() == null) {
                String repoKey = ((RepoDescriptor) diff.getNewValue()).getKey();
                registerRepository(repoKey, true);
            }
            // deleted repo
            if (diff.getOldValue() instanceof RepoDescriptor && diff.getNewValue() == null) {
                String repoKey = ((RepoDescriptor) diff.getOldValue()).getKey();
                deRegisterRepository(repoKey);
            }
        }
    }

    @Override
    public void register(String repoKey, QueueConfiguration configuration) {
        createQueue(repoKey, false, configuration);
    }

    @Override
    public void registerRepository(String repoKey) {
        registerRepository(repoKey, true);
    }

    /**
     * This method is called in the init phase for all active repositories and when creating a new repository.
     *
     * @param newRepository determine if registering an existing repository or a new one
     */
    private void registerRepository(String repoKey, boolean newRepository) {
        if (registrars.isEmpty()) {
            //nothing to do if no registrars
            return;
        }
        RepoDescriptor repoDescriptor = repositoryService.repoDescriptorByKey(repoKey);
        RepoType repoType = repoDescriptor.getType();
        RepoDetailsType detailsType = getRepoDetailsType(repoDescriptor);
        //collect all configurations for a repoKey
        List<QueueConfiguration> configurations = registrars.stream()
                .filter(r -> isRelevant(r, repoType, detailsType))
                .map(r -> r.getConfigurations(repoKey))
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        for (QueueConfiguration queueConfiguration : configurations) {
            createQueue(repoKey, newRepository, queueConfiguration);
        }
    }

    private void createQueue(String repoKey, boolean newRepository, QueueConfiguration queueConfiguration) {
        NodeEventCursorType cursorType = queueConfiguration.getCursorType();
        if (!cursorType.isEventIdType()) {
            throw new IllegalStateException(
                    String.format("cannot accept cursorType %s markerType must be EventId", cursorType));
        }
        //create the new queue and register it
        PersistentQueue persistentQueue = new PersistentQueueImpl(queueConfiguration, repoKey);
        try {
            readWriteLock.writeLock().lock();
            RepoAndCursor key = new RepoAndCursor(repoKey, cursorType);
            //before inserting to the queue - check we dont have the same name already:
            //can't have multiple queues with same name
            //e.g: only 1 ("my-local-maven-repo",IncrementalIndex) queue
            //if we want more than 1 we must use different variants
            //e.g: ("my-local-maven-repo",Replication,var1) queue
            //e.g: ("my-local-maven-repo",Replication,var2) queue
            //this is because a repoKey a variant and a cursorType defines the name of the cursor in the queue.
            //we cannot have 2 different queues working on the same cursor!
            Collection<PersistentQueue> persistentQueues = queues.get(key);
            if (persistentQueues.stream().anyMatch(q -> q.getName().equals(persistentQueue.getName()))) {
                throw new IllegalStateException(
                        String.format("found queue with same name %s", persistentQueue.getName()));
            }
            // Create the marker
            persistentQueue.createMarkerIfNotExists();
            if (newRepository) {
                //whenever a queue for a new repository is created we want to make sure to reset any old marker if
                //exist. The reset method takes care of the cursor and set it to the last event of the queue.
                //This can prevent problems if the original reset(true) call for this queue failed during triggered
                //if we don't do this we could start a new queue with leftover events of a previous queue with same
                //name.
                persistentQueue.reset();
            }
            // Store the qq
            queues.put(key, persistentQueue);
        } finally {
            readWriteLock.writeLock().unlock();
        }
        log.info("Queue created, name: {}, repo: {}", persistentQueue.getName(), repoKey);
    }

    private boolean isRelevant(QueueConfigurationRegistrar r, RepoType repoType, RepoDetailsType detailsType) {
        RepoType requestedRepoType = r.getRepoType();
        RepoDetailsType requestedDetailsType = r.getRepoDetailsType();
        boolean repoTypeMatch = requestedRepoType == null || requestedRepoType.equals(repoType);
        boolean repoDetailsTypeMatch = requestedDetailsType == null || requestedDetailsType.equals(detailsType);
        return repoTypeMatch && repoDetailsTypeMatch;
    }

    @NotNull
    // TODO: 07/07/2020 remove this method once RepoDetailsType will be moved to RepoDescriptor module and
    //  getRepoDetailsType() will introduce on the RepoDescriptor interface
    private RepoDetailsType getRepoDetailsType(RepoDescriptor repoDescriptor) {
        RepoDetailsType detailsType;
        if (repoDescriptor instanceof LocalRepoDescriptor) {
            detailsType = RepoDetailsType.LOCAL;
        } else if (repoDescriptor instanceof RemoteRepoDescriptor) {
            detailsType = RepoDetailsType.REMOTE;
        } else if (repoDescriptor instanceof VirtualRepoDescriptor) {
            detailsType = RepoDetailsType.VIRTUAL;
        } else if (repoDescriptor instanceof FederatedRepoDescriptor) {
            detailsType = RepoDetailsType.FEDERATED;
        } else {
            throw new UnsupportedOperationException("this repository descriptor is not supported yet");
        }
        return detailsType;
    }

    @Override
    public void deRegister(String repoKey, NodeEventCursorType type, String variant) {
        findQueue(repoKey, type, variant).ifPresentOrElse(queue -> {
            try {
                readWriteLock.writeLock().lock();
                queues.remove(new RepoAndCursor(repoKey, type), queue);
                log.info("Queue removed, name: {}, repo: {}", queue.getName(), repoKey);
            } finally {
                readWriteLock.writeLock().unlock();
            }
            queue.shutdown();
            queue.reset(repoKey, true);
        }, () -> log.debug("Cannot found a queue for {} {} {}", repoKey, type, variant));
    }

    @Override
    public void deRegisterRepository(String repoKey) {
        //this method will be called when deleting a repository
        log.debug("Removing all queues for repo {}", repoKey);
        List<PersistentQueue> removedQueues;
        try {
            readWriteLock.writeLock().lock();
            //for each relevant log types try to remove from the map
            //collect the removed queues
            removedQueues = getEventIdLogTypes().stream()
                    .map(type -> queues.removeAll(new RepoAndCursor(repoKey, type)))
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
        } finally {
            readWriteLock.writeLock().unlock();
        }
        if (!removedQueues.isEmpty()) {
            log.info("Unregistering {} queues for repo {}", removedQueues.size(), repoKey);
            //call shutdown on all removed queues - signals all jobs to stop
            removedQueues.forEach(PersistentQueue::shutdown);
            //call the reset on all removed queues - make sure marker is deleted
            //deleting the marker will allow the events cleanup job to do his work.
            removedQueues.forEach(queue -> queue.reset(repoKey, true));
        } else {
            log.debug("No queues found for {} ", repoKey);
        }
    }

    @Override
    public Collection<PersistentQueue> getQueues(String repoKey, NodeEventCursorType cursorType) {
        try {
            readWriteLock.readLock().lock();
            //Wrap the result in a new List to avoid concurrent modification exception
            return Lists.newArrayList(queues.get(new RepoAndCursor(repoKey, cursorType)));
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public Collection<PersistentQueue> getQueues(Set<String> reposKey) {
        try {
            readWriteLock.readLock().lock();
            return queues.entries().stream()
                    .filter(entry -> reposKey.contains(entry.getKey().getRepoKey()))
                    .map(Map.Entry::getValue)
                    .collect(Collectors.toList());
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public Collection<PersistentQueue> getAllQueues() {
        try {
            readWriteLock.readLock().lock();
            return queues.values();
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public boolean isRegistered(String repoKey, NodeEventCursorType type, String variant) {
        return findQueue(repoKey, type, variant).isPresent();
    }

    @Override
    public QueueOperations getQueueOperations(String repoKey, NodeEventCursorType type, String variant) {
        return findQueue(repoKey, type, variant).orElseThrow(() -> new RuntimeException(
                String.format("Couldn't find Queue for %s %s %s", type, repoKey, variant)));
    }

    private Optional<PersistentQueue> findQueue(String repoKey, NodeEventCursorType type, String variant) {
        try {
            readWriteLock.readLock().lock();
            Collection<PersistentQueue> persistentQueues = queues.get(new RepoAndCursor(repoKey, type));
            return persistentQueues.stream()
                    .filter(q -> q.getQueueConfiguration().getVariant().equals(variant))
                    .findFirst();
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public void destroy() {
        readWriteLock.writeLock().lock();
        try {
            queues.values().forEach(PersistentQueue::shutdown);
            queues.clear();
        } finally {
            readWriteLock.writeLock().unlock();
        }
    }
}
