/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.schedule;

import com.google.common.collect.Sets;
import lombok.experimental.Delegate;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.common.ConstantValues;
import org.artifactory.concurrent.LockingException;
import org.artifactory.storage.db.locks.service.DbLocksService;
import org.jfrog.common.concurrency.ConflictsGuard;

import java.util.Arrays;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static java.util.concurrent.TimeUnit.SECONDS;
import static org.artifactory.concurrent.NonStandardLockCategory.CLUSTER_SINGLETON;

/**
 * @author Mor Merhav
 */
@Slf4j
public class HaQuartzTask implements InternalTask {

    public static final String STOP_MARKER_PREFIX = "stop_marker_";
    public static final String PAUSE_MARKER_PREFIX = "pause_marker_";
    private final String stopMarkerLockPrefix;
    private final String pauseMarkerLockPrefix;
    private ConflictsGuard<String> conflictsGuard;
    private DbLocksService dbLocksService;
    private Set<String> markerLocks;
    @Delegate(excludes = ExcludedMethods.class)
    private QuartzTask quartzTask;

    HaQuartzTask(QuartzTask quartzTask, ConflictsGuard<String> conflictsGuard, DbLocksService dbLocksService) {
        this.quartzTask = quartzTask;
        this.conflictsGuard = conflictsGuard;
        this.dbLocksService = dbLocksService;
        this.stopMarkerLockPrefix = STOP_MARKER_PREFIX + getMyJobCommand().jobName() + "_";
        this.pauseMarkerLockPrefix = PAUSE_MARKER_PREFIX + getMyJobCommand().jobName() + "_";
        this.markerLocks = Sets.newConcurrentHashSet();
    }

    @Override
    public void pause(boolean wait, String pausingTaskToken) {
        pause(wait);
        if (pausingTaskToken != null) {
            acquireMarkerLock(pauseMarkerLockPrefix + pausingTaskToken);
        }
    }

    @Override
    public void stop(boolean wait, String stoppingTaskToken) {
        stop(wait);
        if (stoppingTaskToken != null) {
            acquireMarkerLock(stopMarkerLockPrefix + stoppingTaskToken);
        }
    }

    @Override
    public boolean resume(String stoppingTaskToken) {
        boolean isScheduled = resume();
        if (stoppingTaskToken != null && !markerLocks.isEmpty()) {
            log.debug("Removing stop/pause marker locks acquired by: {}", stoppingTaskToken);
            Set<String> markerLocksByStoppingTask =
                    markerLocks.stream()
                            .filter(markerLock -> markerLock.contains(stoppingTaskToken))
                            .collect(Collectors.toUnmodifiableSet());
            markerLocksByStoppingTask.forEach(conflictsGuard::unlock);
            markerLocks.removeAll(markerLocksByStoppingTask);
        }
        return isScheduled;
    }

    @Override
    public boolean started() {
        return quartzTask.started() && !isStoppedByMarkerLock();
    }

    @Override
    public boolean isRunningOnCluster() {
        return isRunning() || isMyClusterSingletonLockClaimed();
    }

    @Override
    public boolean processActiveOnCluster() {
        return processActive() || isRunningOnCluster();
    }

    @Override
    public boolean blockIfPausedAndShouldBreak() {
        boolean shouldStop = quartzTask.blockIfPausedAndShouldBreak();
        if (shouldStop) {
            return true;
        }

        int tries = ConstantValues.taskCompletionLockTimeoutRetries.getInt();
        long timeToPauseMillis = TimeUnit.SECONDS.toMillis(ConstantValues.taskPauseTimeoutSecs.getLong());

        while (isPausedByMarkerLock()) {
            log.debug("Entering paused wait for: {} milliseconds on: {}", timeToPauseMillis, this);
            try {
                Thread.sleep(timeToPauseMillis);
            } catch (InterruptedException e) {
                log.warn("Interrupted while trying to pause task:{}", e.getMessage());
                Thread.currentThread().interrupt();
            }
            tries--;
            if (tries <= 0) {
                throw new LockingException("Task " + this + " paused for more than " +
                        ConstantValues.taskCompletionLockTimeoutRetries.getInt() + " times.");
            }
            log.trace("Job {} waited on pause by marker for {} seconds", this, timeToPauseMillis);
        }

        return isStoppedByMarkerLock();
    }

    private boolean isMyClusterSingletonLockClaimed() {
        return dbLocksService.isLocked(CLUSTER_SINGLETON.getName(), getTaskLockKey());
    }

    private String getTaskLockKey() {
        return getTaskLockKeyIncludingKeyAttributes(Arrays.asList(getType().getAnnotation(JobCommand.class)
                .keyAttributes()));
    }

    private void acquireMarkerLock(String uniqueMarkerLock) {
        try {
            if (!conflictsGuard.isLocked(uniqueMarkerLock)) {
                log.debug("Attempting stop task by acquiring stop marker lock: {}", uniqueMarkerLock);
                conflictsGuard.tryToLock(uniqueMarkerLock, 0, SECONDS);
                log.debug("Successfully stopped task by acquiring stop marker lock: {}", uniqueMarkerLock);
                markerLocks.add(uniqueMarkerLock);
            }
        } catch (InterruptedException e) {
            log.error("Interrupting while trying stop task by acquiring marker lock: {}", uniqueMarkerLock, e);
            Thread.currentThread().interrupt();
        }
    }

    private boolean isStoppedByMarkerLock() {
        boolean isStoppedByMarkerLock;
        if (isUniqueKeyAttributesTask()) {
            isStoppedByMarkerLock =
                    dbLocksService.isLockedByPrefixExcluding(CLUSTER_SINGLETON.getName(), stopMarkerLockPrefix,
                            getType().getSimpleName(), true);
        } else {
            isStoppedByMarkerLock = dbLocksService.isLockedByPrefixExcluding(
                    CLUSTER_SINGLETON.getName(), stopMarkerLockPrefix,
                    stopMarkerLockPrefix + quartzTask.getToken(), false);
        }
        if (isStoppedByMarkerLock) {
            log.debug("Stop marker lock found for job with token: {}", getToken());
        }
        return isStoppedByMarkerLock;
    }

    private boolean isPausedByMarkerLock() {
        boolean isPausedByMarkerLock = dbLocksService
                .isLockedByPrefixExcluding(
                        CLUSTER_SINGLETON.getName(), pauseMarkerLockPrefix,
                        pauseMarkerLockPrefix + quartzTask.getToken(), false);
        if (isPausedByMarkerLock) {
            log.debug("Pause marker lock found for job with token: {}", getToken());
            return true;
        }
        return false;
    }

    private boolean isUniqueKeyAttributesTask() {
        return Arrays.stream(getMyJobCommand().keyAttributes())
                .anyMatch(keyAttribute -> quartzTask.getAttributeMap().containsKey(keyAttribute));
    }

    private JobCommand getMyJobCommand() {
        return this.getType().getAnnotation(JobCommand.class);
    }

    @SuppressWarnings("unused")
    private interface ExcludedMethods {

        void stop(boolean wait, String stoppingTaskToken);

        boolean resume(String stoppingTaskToken);

        void pause(boolean wait, String pausingTaskToken);

        boolean isRunningOnCluster();

        boolean processActiveOnCluster();

        boolean started();

        boolean blockIfPausedAndShouldBreak();

    }
}