/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.work.queue;

import com.google.common.collect.Lists;
import org.artifactory.api.repo.AsyncWorkQueueCustomExecutor;
import org.artifactory.common.ConstantValues;
import org.artifactory.schedule.ArtifactoryConcurrentExecutor;
import org.jfrog.config.utils.IndependentCondition;
import org.springframework.scheduling.concurrent.ConcurrentTaskExecutor;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.context.ContextHelper;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.testng.Assert;
import org.testng.annotations.Test;

import java.lang.reflect.Method;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static org.fest.assertions.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.fail;

/**
 * Unit tests for the {@link WorkQueueImpl}.
 *
 * @author Yossi Shaul
 */
@Test
public class WorkQueueTest extends WorkQueueTestBase {

    public void addWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>(QUEUE_NAME, 5, null, mock(ConflictsGuard.class));

        assertThat(wq.availablePermits()).isEqualTo(5);

        wq.updateWorkers(10);

        assertThat(wq.availablePermits()).isEqualTo(10);

    }

    public void freeWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>(QUEUE_NAME, 5, null, mock(ConflictsGuard.class));

        assertThat(wq.availablePermits()).isEqualTo(5);

        wq.updateWorkers(3);

        assertThat(wq.availablePermits()).isEqualTo(3);

    }

    @Test(expectedExceptions = InterruptedException.class)
    public void freeUsedWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>(QUEUE_NAME, 5, null, mock(ConflictsGuard.class));

        assertThat(wq.availablePermits()).isEqualTo(5);

        wq.tryAcquire();
        wq.tryAcquire();

        assertThat(wq.availablePermits()).isEqualTo(3);

        new IndependentCondition().waitForConditionOnCall(() -> false, TimeUnit.SECONDS, 5L);

        wq.updateWorkers(1);

        fail();

    }

    public void executeAcquired() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq = spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, mock(ConflictsGuard.class)));
        ConcurrentTaskExecutor executor = mock(ConcurrentTaskExecutor.class);
        wq.execute(executor);

        verify(wq, times(1)).tryAcquire();
        verify(wq, never()).release();

        assertThat(wq.availablePermits()).isEqualTo(0);
        verify(executor, times(1)).submit(any(Callable.class));
    }

    public void executeNotAcquired() throws Exception {
        WorkQueueImpl<IntegerWorkItem> wq = spy(new WorkQueueImpl<>(QUEUE_NAME, 0, null, mock(ConflictsGuard.class)));

        ConcurrentTaskExecutor executor = mock(ConcurrentTaskExecutor.class);
        wq.execute(executor);

        verify(wq, times(1)).tryAcquire();
        verify(wq, never()).release();

        assertThat(wq.availablePermits()).isEqualTo(0);
        verify(executor, never()).submit(any(Callable.class));
    }

    public void executeRejected() throws Exception {
        WorkQueueImpl<IntegerWorkItem> wq = spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, mock(ConflictsGuard.class)));
        ConcurrentTaskExecutor executor = mock(ConcurrentTaskExecutor.class);
        when(executor.submit(any(Callable.class))).thenThrow(new RejectedExecutionException());

        wq.execute(executor);

        verify(wq, times(1)).tryAcquire();
        verify(wq, times(1)).release();

        assertThat(wq.availablePermits()).isEqualTo(1);
        verify(executor, times(1)).submit(any(Callable.class));
    }


    public void testReExecute() throws Exception {

        AsyncWorkQueueCustomExecutor customExecutor = AsyncWorkQueueCustomExecutor.builder()
                .threadsCount(() -> 1)
                .repoKey("repo1")
                .nameSuffix("suffix")
                .executor(mock(ConcurrentTaskExecutor.class))
                .build();
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, customExecutor, mock(ConflictsGuard.class)));

        doAnswer(invocation -> false).when(wq).shouldPromote(any(Instant.class));
        when(wq.getQueueSize()).thenReturn(1);

        wq.doJobs();

        verify(wq, times(1)).execute(any(ConcurrentTaskExecutor.class));
    }

    public void testShouldPromote() throws Exception {
        AsyncWorkQueueCustomExecutor customExecutor = AsyncWorkQueueCustomExecutor.builder()
                .threadsCount(() -> 1)
                .repoKey("repo1")
                .nameSuffix("suffix")
                .executor(mock(ConcurrentTaskExecutor.class))
                .build();
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, customExecutor, mock(ConflictsGuard.class)));

        assertThat(wq.shouldPromote(Instant.now())).isTrue();
    }

    public void testShouldPromoteTimeElapsedNoPendingTasks() throws Exception {
        ArtifactoryConcurrentExecutor artifactoryConcurrentExecutor = mock(ArtifactoryConcurrentExecutor.class);
        ConcurrentTaskExecutor concurrentTaskExecutor = spy(new ConcurrentTaskExecutor(artifactoryConcurrentExecutor));
        AsyncWorkQueueCustomExecutor customExecutor = AsyncWorkQueueCustomExecutor.builder()
                .threadsCount(() -> 1)
                .repoKey("repo1")
                .nameSuffix("suffix")
                .executor(concurrentTaskExecutor)
                .build();
        when(artifactoryConcurrentExecutor.getPendingTasks()).thenReturn(0);

        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, customExecutor, mock(ConflictsGuard.class)));

        Instant start = Instant.now().minusSeconds(ConstantValues.workQueueMaxExecutionSec.getLong() + 1);
        assertThat(wq.shouldPromote(start)).isTrue();
    }

    public void testShouldPromoteTimeElapsedAndPendingTasks() throws Exception {
        ArtifactoryConcurrentExecutor artifactoryConcurrentExecutor = mock(ArtifactoryConcurrentExecutor.class);
        ConcurrentTaskExecutor concurrentTaskExecutor = spy(new ConcurrentTaskExecutor(artifactoryConcurrentExecutor));
        AsyncWorkQueueCustomExecutor customExecutor = AsyncWorkQueueCustomExecutor.builder()
                .threadsCount(() -> 1)
                .repoKey("repo1")
                .nameSuffix("suffix")
                .executor(concurrentTaskExecutor)
                .build();
        when(artifactoryConcurrentExecutor.getPendingTasks()).thenReturn(1);

        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, null, customExecutor, mock(ConflictsGuard.class)));

        Instant start = Instant.now().minusSeconds(ConstantValues.workQueueMaxExecutionSec.getLong() + 1);
        assertThat(wq.shouldPromote(start)).isFalse();
    }

    public void offerTheSameWorkConcurrently() throws InterruptedException, NoSuchMethodException {
        List<AtomicInteger> counts = Collections.nCopies(10, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        Method method = counter.getClass().getMethod("count", IntegerWorkItem.class);
        ConflictsGuard guard = ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .getConflictsGuard("Test Queue");
        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>("Test Queue", 1, counter, guard);
        ExecutorService executorService = Executors.newFixedThreadPool(20);
        for (int i = 0; i < 10; i++) {
            for (int j = 0; j < 10000; j++) {
                wq.offerWork(new IntegerWorkItem(i), method);
                executorService.submit(wq::doJobs);
            }
        }

        executorService.shutdown();
        executorService.awaitTermination(1500, TimeUnit.SECONDS);
        // We expect each job to be executed at least once
        counts.forEach(count -> assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }

    public void offerWorkWithNWorkers() throws InterruptedException, NoSuchMethodException {
        List<AtomicInteger> counts = Collections.nCopies(1000, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        Method method = counter.getClass().getMethod("count", IntegerWorkItem.class);
        ConflictsGuard guard = ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .getConflictsGuard("Test Queue");
        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>("Test Queue", 4, counter, guard);

        ExecutorService executorService = Executors.newFixedThreadPool(20);
        for (int i = 0; i < 10; i++) {
            for (int j = 0; j < 10; j++) {
                wq.offerWork(new IntegerWorkItem(i), method);
                executorService.submit(wq::doJobs);
            }
        }

        executorService.shutdown();
        executorService.awaitTermination(1500, TimeUnit.SECONDS);
        // We expect each job to be executed at least once
        counts.forEach(count -> assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }

    // When waiting for item (blockUntilFinished = true), the item we are waiting for can either be pending, or
    //promoted state.
    // This tests a bug that when the item is promoted, we never actually waited for it to complete
    public void waitUntilWhenItemIsPromotedTest() throws NoSuchMethodException, InterruptedException {
        SleepingDummyService dummyService = new SleepingDummyService();
        Method method = dummyService.getClass().getMethod("serve", DummyWorkItem.class);
        ConflictsGuard guard = ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .getConflictsGuard("Test Queue");
        WorkQueueImpl<DummyWorkItem> queue = new WorkQueueImpl<>("Test Queue", 4, dummyService, guard);
        ExecutorService executorService = Executors.newFixedThreadPool(10);
        for (int i = 0; i < 20; i++) {
            DummyWorkItem workItem = new DummyWorkItem("x" + i);
            queue.offerWork(workItem, method);
            if (queue.availablePermits() != 0) {
                // there's a small chance we will miss event (especially if the work queue contain only single worker)
                executorService.submit(queue::doJobs);
            }
            System.out.println("Waiting for workItem number " + i);
            queue.waitForItemDone(workItem);
            Assert.assertTrue(workItem.isProcessed(), "Request num " + i + " failed");
        }
        executorService.shutdown();
        executorService.awaitTermination(1500, TimeUnit.SECONDS);
    }

    public void waitUntilTest() throws InterruptedException, ExecutionException, NoSuchMethodException {
        List<AtomicInteger> counts = Collections.nCopies(10, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        Method method = counter.getClass().getMethod("count", IntegerWorkItem.class);
        ConflictsGuard guard = ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .getConflictsGuard("Test Queue");
        WorkQueueImpl<IntegerWorkItem> wq = new WorkQueueImpl<>("Test Queue", 4, counter, guard);

        ExecutorService executorService = Executors.newFixedThreadPool(10);
        ExecutorService testExecutor = Executors.newFixedThreadPool(10);
        ArrayList<Future> futures = Lists.newArrayList();
        for (int i = 0; i < 5; i++) {
            futures.add(testExecutor.submit(() -> {
                for (int j = 0; j < 10000; j++) {
                    IntegerWorkItem workItem = new IntegerWorkItem((int) (Math.random() * 10d));
                    wq.offerWork(workItem, method);
                    if (wq.tryAcquire()) {
                        // there's a small chance we will miss event (especially if the work queue contain only
                        //single worker)
                        executorService.submit(wq::doJobs);
                    }
                    wq.waitForItemDone(workItem);
                }
                return null;
            }));
        }
        for (Future future : futures) {
            future.get();
        }
        testExecutor.shutdown();
        testExecutor.awaitTermination(1500, TimeUnit.SECONDS);
        executorService.shutdown();
        executorService.awaitTermination(1500, TimeUnit.SECONDS);
        // We expect each job to be executed at least once
        counts.forEach(count -> assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }
}
