/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security;

import com.google.common.collect.Lists;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.api.security.access.CreatedTokenInfo;
import org.artifactory.api.security.access.UserTokenSpec;
import org.artifactory.common.ConstantValues;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.props.auth.OauthManager;
import org.artifactory.security.props.auth.model.AuthenticationModel;
import org.artifactory.security.props.auth.model.OAuth2TokenModel;
import org.artifactory.security.props.auth.model.OauthModel;
import org.artifactory.security.props.auth.model.TokenKeyValue;
import org.artifactory.util.CollectionUtils;
import org.artifactory.util.date.DateUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Iterator;
import java.util.Set;

/**
 * @author Chen  Keinan
 */
@Component
public class LoginHandlerImpl implements LoginHandler {
    private static final Logger log = LoggerFactory.getLogger(LoginHandlerImpl.class);

    private final OauthManager oauthManager;
    private final UserGroupService userGroupService;
    private final AccessService accessService;

    @Autowired
    public LoginHandlerImpl(OauthManager oauthManager, UserGroupService userGroupService,
            AccessService accessService) {
        this.oauthManager = oauthManager;
        this.userGroupService = userGroupService;
        this.accessService = accessService;
    }

    @Override
    public OauthModel doBasicAuthWithDb(String username) {
        return doBasicAuthWithDb(username, false);
    }

    @Override
    public OauthModel getOrCreateOauth2Token(String username) {
        return doBasicAuthWithDb(username, true);
    }

    private OauthModel doBasicAuthWithDb(String username, boolean isOAuth2) {
        TokenKeyValue tokenKeyValue = oauthManager.getToken(username, true);// Use cache for Docker & Conan
        if (tokenKeyValue == null) {
            tokenKeyValue = oauthManager.createToken(username);
        }
        boolean externalUserToken = false;
        if (tokenKeyValue == null) {
            log.debug(
                    "could not create and persist token for authenticated user: '{}', storing generated token in " +
                            "shared cache.",
                    username);
            tokenKeyValue = generateToken(username);
            if (tokenKeyValue == null) {
                throw new RuntimeException("failed to generate token for authenticated user: '" + username + "'");
            }
            externalUserToken = true;
        }
        if (isOAuth2) {
            OAuth2TokenModel oAuth2TokenModel = new OAuth2TokenModel(tokenKeyValue.getToken());
            if (externalUserToken) {
                oAuth2TokenModel.setExpiresIn(ConstantValues.genericTokensCacheIdleTimeSecs.getInt());
            }
            return oAuth2TokenModel;
        }
        AuthenticationModel oauthModel = new AuthenticationModel(tokenKeyValue.getToken(),
                DateUtils.formatBuildDate(System.currentTimeMillis()));
        if (externalUserToken) {
            oauthModel.setExpiresIn(ConstantValues.genericTokensCacheIdleTimeSecs.getInt());
        }
        return oauthModel;
    }

    public TokenKeyValue generateToken(String userName) {
        TokenKeyValue token = null;
        CreatedTokenInfo createdTokenInfo;
        String key = "accesstoken";
        try {
            UserInfo userInfo = userGroupService.currentUser();
            String scope = getScope(userInfo);
            UserTokenSpec tokenSpec = UserTokenSpec.create(userName)
                    .expiresIn(ConstantValues.genericTokensCacheIdleTimeSecs.getLong())
                    .refreshable(false)
                    .scope(Lists.newArrayList(scope));
            createdTokenInfo = accessService.createToken(tokenSpec);
            token = new TokenKeyValue(key, createdTokenInfo.getTokenValue());
        } catch (Exception e) {
            log.debug("Failed generating token for user: '{}' with key '{}'. {}", userName, key, e.getMessage());
            log.trace("Failed generating token.", e);
        }
        return token;
    }

    String getScope(UserInfo userInfo) {
        StringBuilder builder = new StringBuilder("member-of-groups:");
        Set<UserGroupInfo> groups = userInfo.getGroups();
        if (CollectionUtils.isNullOrEmpty(groups)) {
            builder.append("*");
        } else {
            Iterator<UserGroupInfo> it = groups.iterator();
            boolean hasNext = it.hasNext();
            while (hasNext) {
                builder.append(it.next());
                if (hasNext = it.hasNext()) {
                    builder.append(",");
                }
            }
        }
        return builder.toString();
    }
}
