/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.rest.resource.token;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.docker.DockerAddon;
import org.artifactory.addon.rest.PathToken;
import org.artifactory.addon.rest.RestAddon;
import org.artifactory.api.security.AuthorizationException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.access.CreatedTokenInfo;
import org.artifactory.api.security.access.TokenInfo;
import org.artifactory.api.security.access.TokenIssuedByOtherServiceException;
import org.artifactory.api.security.access.TokenNotFoundException;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.common.ConstantValues;
import org.artifactory.rest.exception.AuthorizationRestException;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.token.exception.TokenRequestException;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.access.token.exception.TokenScopeException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.security.acls.model.NotFoundException;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.security.RolesAllowed;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.FormParam;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static javax.ws.rs.core.MediaType.APPLICATION_FORM_URLENCODED;
import static javax.ws.rs.core.MediaType.APPLICATION_JSON;
import static javax.ws.rs.core.MediaType.TEXT_PLAIN;
import static org.apache.commons.lang.StringUtils.isNotBlank;
import static org.artifactory.api.security.AuthorizationService.ROLE_ADMIN;
import static org.artifactory.api.security.AuthorizationService.ROLE_USER;
import static org.artifactory.rest.resource.token.TokenResource.PATH_ROOT;
import static org.artifactory.security.token.TokenResponseErrorCode.InvalidGrant;
import static org.artifactory.security.token.TokenResponseErrorCode.InvalidRepoPath;
import static org.artifactory.security.token.TokenResponseErrorCode.InvalidRequest;
import static org.artifactory.security.token.TokenResponseErrorCode.InvalidScope;
import static org.artifactory.security.token.model.TokenUtils.assertSupportedTokenSpecForCreate;
import static org.artifactory.security.token.model.TokenUtils.audienceToList;
import static org.artifactory.security.token.model.TokenUtils.createTokenSpec;
import static org.artifactory.security.token.model.TokenUtils.expiresInOrDefault;
import static org.artifactory.security.token.model.TokenUtils.requireUsername;
import static org.jfrog.access.util.TokenScopeUtils.scopeToList;

/**
 * @author Yinon Avraham
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Path(PATH_ROOT)
@RolesAllowed({ROLE_ADMIN, ROLE_USER})
public class TokenResource {

    public static final String PATH_ROOT = "security/token";
    private static final Logger log = LoggerFactory.getLogger(TokenResource.class);
    private static final Set<String> REQUIRED_REFRESH_TOKEN_REQUEST_PARAMS = ImmutableSet
            .of("grant_type", "refresh_token", "access_token");
    private static final Set<String> OPTIONAL_REFRESH_TOKEN_REQUEST_PARAMS = ImmutableSet.of("expires_in");

    @Autowired
    private AccessService accessService;

    @Autowired
    private AuthorizationService authorizationService;

    @Autowired
    private AddonsManager addonsManager;

    @Context
    private HttpServletRequest request;

    @POST
    @Consumes(APPLICATION_FORM_URLENCODED)
    @Produces(APPLICATION_JSON)
    public Response createOrRefreshToken(
            @FormParam("grant_type") String grantTypeName,
            @FormParam("username") String username,
            @FormParam("refresh_token") String refreshToken,
            @FormParam("access_token") String accessToken,
            @FormParam("scope") String scope,
            @FormParam("expires_in") Long expiresIn,
            @FormParam("refreshable") @DefaultValue("false") boolean refreshable,
            @FormParam("audience") String audience) {
        GrantType grantType = requireGrantType(grantTypeName);
        assertPermittedOperation(grantType);
        long expiryDefault = getExpiryDefault(accessToken, StringUtils.isNotBlank(refreshToken));
        long targetExpiresIn = expiresInOrDefault(expiresIn, expiryDefault);
        List<String> scopeTokens;
        try {
            scopeTokens = scopeToList(scope);
        } catch (TokenScopeException e) {
            throw new TokenRequestException(InvalidScope, e.getMessage());
        }
        List<String> audienceList = audienceToList(audience);
        try {
            TokenSpec tokenSpec = createTokenSpec(username)
                    .scope(scopeTokens)
                    .expiresIn(targetExpiresIn)
                    .refreshable(refreshable)
                    .audience(audienceList);
            switch (grantType) {
                case ClientCredentials:
                    assertSupportedTokenSpecForCreate(tokenSpec,
                            accessService.getArtifactoryServiceId().getFormattedName());
                    requireUsername(username);
                    return createToken(tokenSpec);
                case RefreshToken:
                    return refreshToken(tokenSpec, accessToken, refreshToken);
                default:
                    throw new GrantTypeNotSupportedException(grantType.getSignature());
            }
        } catch (IllegalArgumentException | AuthorizationException e) {
            throw new TokenRequestException(InvalidRequest, e);
        } catch (NotFoundException e) {
            throw new TokenRequestException(InvalidRepoPath, e);
        }
    }

    private long getExpiryDefault(String accessToken, boolean refreshToken) {
        if (refreshToken && StringUtils.isNotBlank(accessToken)) {
            try {
                JwtAccessToken jwtAccessToken = accessService.parseToken(accessToken);
                Long expirationTime = jwtAccessToken.getExpirationTime();
                if (expirationTime != null) {
                    return TimeUnit.MILLISECONDS.toSeconds(expirationTime - jwtAccessToken.getIssuedAt());
                }
            } catch (IllegalArgumentException e) {
                log.debug("Unable to parse token, assuming default expiration.", e);
            }
        }
        return ConstantValues.accessTokenExpiresInDefault.getLong();
    }

    @POST
    @Path("filename")
    @Consumes(APPLICATION_JSON)
    @Produces(APPLICATION_JSON)
    public Response tokenForFilename(FilenameRequest filenameRequest) {
        if (ConstantValues.accessTokenForFileLimitedForAnonymous.getBoolean()) {
            assertPermittedOperation(GrantType.ClientCredentials);
        }
        DockerAddon dockerAddon = addonsManager.addonByType(DockerAddon.class);
        RestAddon restAddon = addonsManager.addonByType(RestAddon.class);
        PathToken pathToken = restAddon.createPathToken(filenameRequest.getRepoKey(), filenameRequest.getFilename(),
                filenameRequest.getExpireIn(), dockerAddon.getBlobsRepoPathCache());
        return Response.ok(pathToken).build();
    }

    @POST
    @Path("path")
    @Consumes(APPLICATION_JSON)
    @Produces(APPLICATION_JSON)
    public Response tokenForPath(FilePathRequest filePathRequest) {
        if (ConstantValues.accessTokenForFileLimitedForAnonymous.getBoolean()) {
            assertPermittedOperation(GrantType.ClientCredentials);
        }
        RestAddon restAddon = addonsManager.addonByType(RestAddon.class);
        PathToken pathToken = restAddon.createPathToken(filePathRequest.getPath(), filePathRequest.getExpiresIn());
        return Response.ok(pathToken).build();
    }

    @RolesAllowed({ROLE_ADMIN})
    @GET
    @Produces(APPLICATION_JSON)
    public Response getTokenInfos() {
        List<TokenInfo> tokens = accessService.getTokenInfos();
        List<TokenInfoModel> result = tokens.stream().map(TokenInfoModel::new).collect(Collectors.toList());
        TokensResponseModel responseModel = new TokensResponseModel();
        responseModel.setTokens(result);
        return Response.ok(responseModel).build();
    }

    private void assertPermittedOperation(@Nonnull GrantType grantType) {
        if (!authorizationService.isAdmin()) {
            switch (grantType) {
                case RefreshToken:
                    if (nonAdminCanRefreshToken()) {
                        return;
                    }
                    break;
                case ClientCredentials:
                    if (nonAdminCanCreateToken()) {
                        return;
                    }
                    break;
                default:
                    throw new GrantTypeNotSupportedException(grantType.getSignature());
            }
            if (!authorizationService.isAuthenticated() || authorizationService.isAnonymous()) {
                throw new WebApplicationException(HttpStatus.SC_UNAUTHORIZED);
            }
            throw new AuthorizationRestException();
        }
    }

    private boolean nonAdminCanCreateToken() {
        return authorizationService.isAuthenticated() && !authorizationService.isAnonymous();
    }

    private boolean nonAdminCanRefreshToken() {
        if (isSimpleRefreshTokenRequestParams()) {
            log.debug("Simple refresh token request - no need to authenticate user, does not require admin");
            return true;
        }
        return false;
    }

    private boolean isSimpleRefreshTokenRequestParams() {
        Set<String> paramNames = Sets.newHashSet(Collections.list(request.getParameterNames()));
        if (!paramNames.containsAll(REQUIRED_REFRESH_TOKEN_REQUEST_PARAMS)) {
            // One of the required params is missing
            return false;
        }
        paramNames.removeAll(REQUIRED_REFRESH_TOKEN_REQUEST_PARAMS);
        for (String param : paramNames) {
            if (!OPTIONAL_REFRESH_TOKEN_REQUEST_PARAMS.contains(param)) {
                // Param is not allowed.
                return false;
            }
        }
        return true;
    }

    /**
     * Revoke a token by the access token or refresh token
     * See <a href="https://tools.ietf.org/html/rfc7009">RFC7009 - Section 2 - Token Revocation</a>
     * Another option is to give the token ID. This option and the first option (by token & hint) are mutually
     * exclusive.
     *
     * @param token         the access token or refresh token to revoke
     * @param tokenTypeHint the token type hint (access token or refresh token)
     * @param tokenId       the ID of the token to revoke
     */
    @POST
    @Consumes(APPLICATION_FORM_URLENCODED)
    @Produces({TEXT_PLAIN, APPLICATION_JSON})
    @Path("revoke")
    @RolesAllowed({ROLE_ADMIN})
    public Response revokeToken(
            @FormParam("token") String token,
            @FormParam("token_type_hint") String tokenTypeHint,
            @FormParam("token_id") String tokenId) {
        RevokeTokenOption option = assertValidRevokeTokenRequest(token, tokenTypeHint, tokenId);
        TokenType tokenType = tokenTypeHint == null ? TokenType.AccessToken : TokenType.fromSignature(tokenTypeHint);
        //The token type hint is ignored for now... Just validate it in case it is provided.
        String successMessage = "Token revoked";
        try {
            switch (option) {
                case Token:
                    accessService.revokeToken(token);
                    break;
                case TokenId:
                    accessService.revokeTokenById(tokenId);
                    break;
                default:
                    log.warn("Unexpected revoke token option: {}", option);
                    throw new TokenRequestException(InvalidRequest, "token or token_id are required");
            }
        } catch (TokenIssuedByOtherServiceException e) {
            throw new TokenRequestException(InvalidRequest, e);
        } catch (TokenNotFoundException e) {
            log.warn("Ignoring revoke token request, token not found.");
            log.debug("Ignoring revoke token request, token not found.", e);
            successMessage = "Token not found";
        }
        return Response.ok(successMessage).type(MediaType.TEXT_PLAIN_TYPE).build();
    }

    private RevokeTokenOption assertValidRevokeTokenRequest(String token, String tokenTypeHint, String tokenId) {
        if (isNotBlank(token)) {
            if (isNotBlank(tokenId)) {
                throw new TokenRequestException(InvalidRequest, "token and token_id are mutually exclusive");
            }
            return RevokeTokenOption.Token;
        }
        if (isNotBlank(tokenId)) {
            return RevokeTokenOption.TokenId;
        }
        throw new TokenRequestException(InvalidRequest, "token or token_id are required");
    }

    private enum RevokeTokenOption {
        Token, TokenId
    }

    private Response refreshToken(@Nonnull TokenSpec tokenSpec, String accessToken, String refreshToken) {
        try {
            requireAccessToken(accessToken);
            requireRefreshToken(refreshToken);
            CreatedTokenInfo createdTokenInfo = accessService.refreshToken(tokenSpec, accessToken, refreshToken);
            TokenResponseModel model = toTokenResponseModel(createdTokenInfo);
            return Response.ok(model).build();
        } catch (TokenNotFoundException | AuthorizationException e) {
            throw new TokenRequestException(InvalidGrant, e);
        } catch (TokenIssuedByOtherServiceException e) {
            throw new TokenRequestException(InvalidRequest, e);
        }
    }

    @Nonnull
    private String requireAccessToken(String accessToken) {
        if (StringUtils.isBlank(accessToken)) {
            throw new TokenRequestException(InvalidGrant, "access token is required.");
        }
        return accessToken;
    }

    @Nonnull
    private String requireRefreshToken(String refreshToken) {
        if (StringUtils.isBlank(refreshToken)) {
            throw new TokenRequestException(InvalidGrant, "refresh token is required.");
        }
        return refreshToken;
    }

    @Nonnull
    private Response createToken(@Nonnull TokenSpec tokenSpec) {
        CreatedTokenInfo createdTokenInfo = accessService.createToken(tokenSpec);
        TokenResponseModel model = toTokenResponseModel(createdTokenInfo);
        return Response.ok(model).build();
    }

    @Nonnull
    private TokenResponseModel toTokenResponseModel(CreatedTokenInfo createdTokenInfo) {
        TokenResponseModel model = new TokenResponseModel();
        model.setAccessToken(createdTokenInfo.getTokenValue());
        model.setExpiresIn(createdTokenInfo.getExpiresIn());
        model.setScope(createdTokenInfo.getScope());
        model.setTokenType(createdTokenInfo.getTokenType());
        model.setRefreshToken(createdTokenInfo.getRefreshToken());
        return model;
    }

    @Nonnull
    private GrantType requireGrantType(@Nullable String grantTypeName) {
        if (grantTypeName == null) {
            return GrantType.ClientCredentials;
        }
        return GrantType.fromSignature(grantTypeName);
    }

}
